"""GluonCV-Torch video action recognition."""
from __future__ import absolute_import
from .actionrec_resnetv1b import *
from .i3d_resnet import *
from .i3d_slow import *
from .slowfast import *
from .r2plus1dv1 import *
from .r2plus1dv2 import *
from .tpn import *
from .ircsnv2 import *
